'''
Configuration for the seleced bodies.
'''

bodies_dict = {'Mercury': ['199', 'darkturquoise'],
               'Venus': ['299', 'darkorchid'],
               'Earth': ['399', 'green'],
               'Mars': ['499', 'maroon'],
               'Jupiter': ['599', 'navy'],
               'STEREO B': ['-235', 'b'],
               'STEREO A': ['-234', 'red'],
               'Solar Orbiter': ['-144', 'dodgerblue'],
               'Parker Solar Probe': ['-96', 'purple'],
               'BepiColombo': ['-121', 'orange']}
