"""
A place for the more generic tests that do not fit in the other categories

Note
----
Before you start the test the PyThea pkg should be in the python path
export PYTHONPATH="${PYTHONPATH}:{top_level_dir_that_pythea_lives}/PyThea"
"""
